/* usbn960x.c
 * Copyright (C) 2006  Benedikt Sauter
 * Copyright (C) 2008  Manuel Stahl
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <avr/pgmspace.h>
#include <stdlib.h>
#include "usbn960x.h"
#include "../../usbn2mc.h"

uint8_t EP0RXBuf[8];
EPInfo EP0rx = { RXD0, RXC0, EPC0, 8, 0, EP_MEM, 0, 0, NULL};
EPInfo EP0tx = { TXD0, TXC0, EPC0, 8, 0, EP_MEM, 0, 0, NULL};

prog_uint8_t _USBN_langDesc[] = { 0x04, STRING, 0x09, 0x04 };
prog_uint8_t *DeviceDescriptor;
prog_uint8_t *ConfigurationDescriptor;
uint8_t USB_numStringDescriptors;
USB_StringDescriptor_t StringDescriptors[10];
FIFOCallback* RX1Callback;

static inline void _USBNReceiveFIFO0(void);
static inline void _USBNTransmitFIFO0(void);
static inline void _USBNDecodeStandardRequest(DeviceRequest_t *req);

// ********************************************************************
// Communicate with the FIFOs
// ********************************************************************


/**
 * Transmit data from RAM to endpoint.
 */
static inline void _USBNDataToFifo(EPInfo* ep) {
	for (uint8_t i = 0; (i < ep->usbnFifoSize) && (ep->bufIndex < ep->bufSize); i++)
		USBNWrite(ep->usbnData, ep->buf[ep->bufIndex++]);
}

/**
 * Transmit data from flash memory to endpoint.
 */
static inline void _USBNDataToFifo_P(EPInfo* ep) {
	for (uint8_t i = 0; (i < ep->usbnFifoSize) && (ep->bufIndex < ep->bufSize); i++) {
		USBNWrite(ep->usbnData, pgm_read_byte(ep->buf + ep->bufIndex));
		ep->bufIndex++;
	}
}

/**
 * Transmit a string descriptor from flash memory to endpoint.
 */
static inline void _USBNStringToFifo_P(EPInfo* ep) {
	uint8_t i = 0;

	if (ep->bufIndex == 0) { // first send descriptor header
		USBNWrite(ep->usbnData, ep->bufSize * 2 + 2); //    length will be double the string length
		USBNWrite(ep->usbnData, STRING); //    descriptor type
		i = 2;
	}
	for (; (i < ep->usbnFifoSize) && (ep->bufIndex < ep->bufSize); i += 2) {
		USBNWrite(ep->usbnData, pgm_read_byte(ep->buf + ep->bufIndex++)); // transmit character
		USBNWrite(ep->usbnData, 0x00); // fill in zero
	}
}

void USBNReceive(EPInfo* ep) {
	uint8_t count = ep->bufSize;
	for (ep->bufIndex = 0; ep->bufIndex < count; ep->bufIndex++)
		ep->buf[ep->bufIndex] = USBNRead(ep->usbnData);

	USBNWrite(ep->usbnCommand, FLUSH);
	USBNWrite(ep->usbnCommand, RX_EN);
}

void USBNTransmit(EPInfo* ep) {
	uint8_t togl = 0;

	if (ep->bufSize == 0 || ep->buf == 0)
		return; // data in buffer?

	if (ep->bufIndex < ep->bufSize) { // more data to send?

		USBNWrite(ep->usbnCommand, FLUSH); //send data to the FIFO TODO: respect space in fifo
		if (ep->bufType == EP_MEM)
			_USBNDataToFifo(ep);
		else if (ep->bufType == EP_FLASH)
			_USBNDataToFifo_P(ep);
		else if (ep->bufType == EP_STRING)
			_USBNStringToFifo_P(ep);

	}

	if (ep->bufIndex >= ep->bufSize)	// was this the last packet?
		togl |= TX_LAST;

	if (ep->DataPid != 0x00)			// toggle PID
		togl |= TX_TOGL;
	ep->DataPid ^= 0x01;
	USBNWrite(ep->usbnCommand, togl + TX_EN);
}



// ********************************************************************
// Interrupt Event Handler
// ********************************************************************


static inline void _USBNNackEvent(void) {
	uint8_t event;
	event = USBNRead(NAKEV);
	// TODO: error handler

	//USBNWrite(RXC1,FLUSH);	//re-enable the receiver
	//USBNWrite(RXC1,RX_EN);	//re-enable the receiver
	/*
	 if (EP0tx.Size > EP0tx.usbnfifo)	  //multi-pkt status stage?
	 {
	 //USBNDebug("flush");
	 USBNWrite(TXC0,FLUSH);	    //flush TX0 and disable
	 //USBNWrite(RXC0,RX_EN);	//re-enable the receiver
	 EP0tx.DataPid	      = 1;
	 }
	 */
}


static inline void _USBNReceiveEvent(void) {
	uint8_t event = USBNRead(RXEV);

	if (event & RX_FIFO0)
		_USBNReceiveFIFO0();
	else if (event & RX_FIFO1 && RX1Callback != 0) // dynamic function call
		RX1Callback(USBNRead(RXS1) & 0x0f);
}


static inline void _USBNTransmitEvent(void) {
	uint8_t event;
	event = USBNRead(TXEV);
	//USBNDebug("tx event\r\n");
	if (event & TX_FIFO0)
		_USBNTransmitFIFO0();
	else {
		USBNDebug("tx event\r\n");
		USBNRead(TXS1); // get transmitter status
		USBNRead(TXS2); // get transmitter status
		USBNRead(TXS3); // get transmitter status
	}
}

static inline void _USBNAlternateEvent(void) {
	uint8_t event;
	event = USBNRead(ALTEV);
	USBNDebug("alt event\r\n");

	if (event & ALT_RESET) {
		USBNWrite(NFSR, RST_ST); // NFS = NodeReset
		USBNWrite(FAR, AD_EN + 0);
		USBNWrite(EPC0, 0x00);
		USBNWrite(TXC0, FLUSH);
		USBNWrite(RXC0, RX_EN); // allow reception
		USBNWrite(NFSR, OPR_ST); // NFS = NodeOperational
		USBNDebug("reset\r\n");
	}
	if (event & ALT_SD3) {
		USBNWrite(ALTMSK, ALT_RESUME + ALT_RESET); // adjust interrupts
		/*
		 USBNWrite(NFSR,SUS_ST);                   // enter suspend state
		 USBNDebug("sd3\r\n");
		 */
	}
	if (event & ALT_RESUME) {
		USBNWrite(ALTMSK, ALT_SD3 + ALT_RESET + ALT_RESUME);
		/*
		 USBNWrite(EPC0,0x00);
		 USBNWrite(RXC0,RX_EN);                    // allow reception
		 USBNWrite(TXC0,FLUSH);
		 USBNWrite(NFSR,OPR_ST);
		 USBNDebug("resume\r\n");
		 */
	}
	if (event & ALT_EOP) {
		USBNDebug("eop\r\n");
	}

}



// ********************************************************************
// Receive and Transmit functions for EP0
// ********************************************************************

/**
 * Receive request on EP0.
 * Called by _USBNReceiveEvent()
 */
static inline void _USBNReceiveFIFO0(void) {
	uint8_t rxstatus;
	DeviceRequest_t req;

	USBNDebug("rx\r\n");

	rxstatus = USBNRead(RXS0);
	if (rxstatus & SETUP_R) {
		EP0rx.buf = (uint8_t*)&req;
		EP0rx.bufSize = sizeof(DeviceRequest_t);
		USBNReceive(&EP0rx);

		USBNWrite(RXC0, FLUSH); // make sure the RX is off
		USBNWrite(TXC0, FLUSH); // make sure the TX is off
		USBNWrite(EPC0, USBNRead(EPC0) & 0x7F); // turn of stall

		// noch ein switch um zu entscheiden obs fuers device, interface endpoint oder andere ist
		switch (req.bmRequestType & 0x60) // decode request type
		{
		  case DO_STANDARD:			// standard request
			_USBNDecodeStandardRequest(&req);
			break;
		  case DO_CLASS:			// class request
			USBNDebug("Class request\n\r");
			USBNDecodeClassRequest(&req);
			break;
		  case DO_VENDOR:			// vendor request
			USBNDebug("Vendor request\n\r");
			USBNDecodeVendorRequest(&req);
			break;
		  default:					// unsupported req type
			USBNDebug("unsupported req type\r\n");
			USBNWrite(EPC0, USBNRead(EPC0) + STALL); // stall the endpoint
			break;
		}
		//the following is done for all setup packets.  Note that if
		//no data was stuffed into the FIFO, the result of the fol-
		//lowing will be a zero-length response.

		// only for stage 2 transfers
		if (req.bmRequestType == 0x00)
			USBNWrite(TXC0, TX_TOGL + TX_EN); //enable the TX (DATA1)
	} else { // if not a setuppacket

		//USBNDebug("error transmit\r\n");
		if (EP0tx.bufSize > EP0tx.usbnFifoSize) { // multi-pkt status stage?
			if ((rxstatus& 0x5F)!=0x10) // length error??
				USBNDebug("length error\r\n");

			EP0tx.bufSize = 0; // exit multi-packet mode
			USBNWrite(TXC0, FLUSH); // flush TX0 and disable
			USBNWrite(RXC0, RX_EN); // re-enable the receiver
		}
	}
}

/**
 * Transmit request on EP0 has finished.
 * Called by _USBNTransmitEvent()
 */
static inline void _USBNTransmitFIFO0(void) {
	uint8_t txstat;

	txstat = USBNRead(TXS0);	// get transmitter status
	USBNDebug("tx");

	if (txstat & TX_DONE) {		// if transmit completed
		USBNWrite(TXC0, FLUSH);		// flush TX0 and disable

		if (txstat & ACK_STAT) {	// ACK received

			if (EP0tx.bufIndex < EP0tx.bufSize)
				USBNTransmit(&EP0tx);	// transmit rest of packet
			else 				// not in multi-packet mode
				USBNWrite(RXC0, RX_EN); // re-enable the receiver

		} else {					// No ACK: this probably means we issued a stall handshake
			USBNDebug("stall handshake ");
			USBNWrite(RXC0, RX_EN); // re-enable the receiver
		}
	}
	// otherwise something must have gone wrong with the previous
	// transmission, or we got here somehow we shouldn't have
	else {
		//USBNDebug("tx0 error\n");
	}
	// we do this stuff for all tx_0 events
}

// ********************************************************************
// standard request from EP0
// ********************************************************************


static inline void _USBNGetDescriptor(DeviceRequest_t *req) {
	uint8_t index = req->wValue;
	uint8_t type = req->wValue >> 8;

	EP0tx.bufIndex = 0;
	EP0tx.DataPid = 1;
	switch (type) {
	case DEVICE:
		USBNDebug("DEVICE DESCRIPTOR\n\r");
		EP0tx.bufType = EP_FLASH;
		EP0tx.bufSize = pgm_read_byte(&DeviceDescriptor[0]);
		EP0tx.buf = DeviceDescriptor;

		// first get descriptor request is
		// always be answered with first 8 uint8_ts of dev descriptor
		if (req->wLength == 0x40)
			EP0tx.bufSize = 8;
		break;
	case CONFIGURATION:
		USBNDebug("CONFIGURATION DESCRIPTOR ");

		// send complete tree
		EP0tx.bufType = EP_FLASH;
		EP0tx.bufSize = req->wLength;
		EP0tx.buf = ConfigurationDescriptor;

		break;
	case STRING:
		if (index > 0) {
			index--;
			EP0tx.bufType = EP_STRING;
			EP0tx.buf = StringDescriptors[index].String;
			EP0tx.bufSize = StringDescriptors[index].bLength;
		} else {
			EP0tx.bufType = EP_FLASH;
			EP0tx.buf = _USBN_langDesc;
			EP0tx.bufSize = 4;
		}
		break;
	}

	USBNTransmit(&EP0tx);
}

static inline void _USBNDecodeStandardRequest(DeviceRequest_t *req) {
	if ((req->bmRequestType & 0x1f) == 0x00) {
		switch (req->bRequest) { // decode request code

#if 0
		case CLR_FEATURE:
		USBNDebug("CLR FEATURE\n\r");
		//_USBNClearFeature(req);
		break;
		case GET_CONFIGURATION:
		//USBNDebug("GET CONFIG\n\r");
		USBNWrite(TXD0,USBNFunctionInfo.ConfigurationIndex);
		break;
#endif
		case GET_DESCRIPTOR:
			USBNDebug("GET DESCRIPTOR\n\r");
			_USBNGetDescriptor(req);
			break;
		case GET_INTERFACE:
			USBNDebug("GET INTERFACE\n\r");
			// TODO: implement
			break;
		case GET_STATUS:
			USBNDebug("GET STATUS\n\r");
			USBNWrite(TXC0,FLUSH);
			USBNWrite(TXD0,1);
			USBNWrite(TXD0,0);
			USBNWrite(TXC0,TX_TOGL + TX_EN); //enable the TX (DATA1)
			break;
		case SET_ADDRESS:
			USBNDebug("SET ADDRESS ");
			USBNWrite(EPC0, DEF);
			USBNWrite(FAR, AD_EN + req->wValue);
			break;
		case SET_CONFIGURATION:
			USBNDebug("SET CONFIGURATION\n\r");
			USBNSetConfiguration(req);
			break;
#if 0
			case SET_FEATURE:
			USBNDebug("SET FEATURE\n\r");
			break;
#endif
		case SET_INTERFACE:
			USBNDebug("SET INTERFACE\n\r");
			//if(EP0rx.Buf[2])
			//   USBNWrite(EPC1,0);      // stall the endpoint
			USBNWrite(TXC0, TX_TOGL + TX_EN); //enable the TX (DATA1)
			break;
		default: // unsupported standard req
			USBNDebug("unsupported standard req\n\r");
			USBNWrite(EPC0, USBNRead(EPC0) + STALL); // stall the endpoint
			break;
		}
	} else { // default request but for interface not for device
		USBNInterfaceRequests(req, &EP0tx);
		USBNTransmit(&EP0tx);
	}
}


// ********************************************************************
// Interrupt Routine for USBN960x
// ********************************************************************

#ifndef USBNInterrupt
#error USBNInterrupt must be defined in usbn2mc.h
#else

USBNInterrupt() {
	uint8_t maev, mask;

	maev = USBNRead(MAEV);

	// TODO: shouldn't this be without else?
	if (maev & RX_EV)
		_USBNReceiveEvent();
	else if (maev & TX_EV)
		_USBNTransmitEvent();
	else if (maev & ALT)
		_USBNAlternateEvent();
	else if (maev & NAK)
		_USBNNackEvent();

	mask = USBNRead(MAMSK);
	USBNWrite(MAMSK, 0x00); // ack irq request
	USBNWrite(MAMSK, mask);
}
#endif
